/*********************************************************************************************************
* Software License Agreement (BSD License)                                                               *
* Author: Sebastien Decugis <sdecugis@nict.go.jp>							 *
*													 *
* Copyright (c) 2008, WIDE Project and NICT								 *
* All rights reserved.											 *
* 													 *
* Redistribution and use of this software in source and binary forms, with or without modification, are  *
* permitted provided that the following conditions are met:						 *
* 													 *
* * Redistributions of source code must retain the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer.										 *
*    													 *
* * Redistributions in binary form must reproduce the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer in the documentation and/or other						 *
*   materials provided with the distribution.								 *
* 													 *
* * Neither the name of the WIDE Project or NICT nor the 						 *
*   names of its contributors may be used to endorse or 						 *
*   promote products derived from this software without 						 *
*   specific prior written permission of WIDE Project and 						 *
*   NICT.												 *
* 													 *
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED *
* WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A *
* PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR *
* ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 	 *
* LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 	 *
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR *
* TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF   *
* ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.								 *
*********************************************************************************************************/

/* Some useful functions for both radius_gw and its sub extensions 
 * These functions are compiled as a static library rg_common.
 */

#define IN_EXTENSION
#define DEFINE_DEBUG_MACRO	rg_common
#include <waaad/waaad.h>

#include "radius_gw.h"

#include <stdlib.h>
#include <string.h>

#ifndef RG_COMMON_VERBO
#define RG_COMMON_VERBO	1
#endif /* RG_COMMON_VERBO */
static int rg_common_verbosity = RG_COMMON_VERBO;

void rg_list_init(struct rg_list * plist)
{
	ASSERT(plist != NULL);
	plist->next = plist;
	plist->prev = plist;
	plist->head = plist;
}

int rg_list_is_empty(struct rg_list * plist)
{
	ASSERT(plist != NULL);
	return plist == plist->next;
}

void rg_list_insert_after(struct rg_list * ref, struct rg_list * item)
{
	ASSERT(ref != NULL);
	ASSERT(item != NULL);
	ASSERT(rg_list_is_empty(item));
	item->next = ref->next;
	item->prev = ref;
	item->head = ref->head;
	ref->next->prev = item;
	ref->next = item;
}

void rg_list_insert_before(struct rg_list * ref, struct rg_list * item)
{
	ASSERT(ref != NULL);
	ASSERT(item != NULL);
	ASSERT(rg_list_is_empty(item));
	item->prev = ref->prev;
	item->next = ref;
	item->head = ref->head;
	ref->prev->next = item;
	ref->prev = item;
}

void rg_list_unlink(struct rg_list * plist)
{
	ASSERT(plist != NULL);
	if (plist->head == plist)
		return;
	plist->next->prev = plist->prev;
	plist->prev->next = plist->next;
	plist->next = plist;
	plist->prev = plist;
	plist->head = plist;
}


rad_t * rg_msg_create_ans(rad_t * req)
{
	rad_t * ans = NULL;
	rad_t * reqcpy = NULL;
	
	CHECK_MALLOC_DO( reqcpy = malloc(sizeof(rad_t)), return NULL );
	CHECK_MALLOC_DO( ans = malloc(sizeof(rad_t)), { free(reqcpy); return NULL; } );
	
	memcpy(reqcpy, req, sizeof(rad_t));
	rg_list_init(&reqcpy->attributes);
	
	memset(ans, 0, sizeof(rad_t));
	rg_list_init(&ans->attributes);
	
	ans->req = reqcpy;
	
	return ans;
}
	

void rg_msg_free(rad_t * msg)
{
	while ( ! rg_list_is_empty( &msg->attributes ) ) {
		struct rg_list * item = msg->attributes.next;
		rg_list_unlink(item);
		free(item);
	}
	free(msg->req);
	free(msg);
}

char * rg_msg_code_str(uint8_t c)
{
	/* Include generated source file, see rebuild_inc.sh script */
#include "rg_utils_codes.inc"
	
	/* fallback */
	return "[unknown]";
}

char * rg_msg_attrtype_str(uint8_t c)
{
	/* Include generated source file, see rebuild_inc.sh script */
#include "rg_utils_attrtype.inc"
	
	/* fallback */
	return "[unknown]";
}

void rg_msg_dump(int level, rad_t * msg)
{
	struct rg_list * attr;
	
	log_debug("------ RADIUS msg dump -------\n");
	log_debug("   id: %02hhx, code: %hhd (%s)\n", msg->identifier, msg->code, rg_msg_code_str(msg->code));
	log_debug(" auth: %02hhx %02hhx %02hhx %02hhx  %02hhx %02hhx %02hhx %02hhx\n",
			msg->authenticator[0], msg->authenticator[1], msg->authenticator[2], msg->authenticator[3], 
			msg->authenticator[4], msg->authenticator[5], msg->authenticator[6], msg->authenticator[7]);
	log_debug("       %02hhx %02hhx %02hhx %02hhx  %02hhx %02hhx %02hhx %02hhx\n",
			msg->authenticator[8], msg->authenticator[9], msg->authenticator[10], msg->authenticator[11], 
			msg->authenticator[12], msg->authenticator[13], msg->authenticator[14], msg->authenticator[15]);
	for (attr = msg->attributes.next; attr != &msg->attributes; attr = attr->next) {
		struct rad_attr * loc = (struct rad_attr *)attr;
		log_debug("    attr: len:%3hhd, type:%02hhx (%s)\n", loc->length, loc->type, rg_msg_attrtype_str(loc->type));
	}
	log_debug("-----------------------------\n");
}


/* Terminate a thread */
int rg_thread_term(pthread_t * th)
{
	int ret = 0;
	void * th_ret = NULL;
	
	CHECK_PARAMS(th);
	
	/* Test if it was already terminated */
	if (*th == (pthread_t)NULL)
		return 0;
	
	/* Cancel the thread if it is still running - ignore error if it was already terminated */
	(void) pthread_cancel(*th);
	
	/* Then join the thread */
	CHECK_POSIX_DO( ret = pthread_join(*th, &th_ret), /* continue */ );
	
	if (th_ret != NULL) {
		TRACE_DEBUG(FULL, "The thread returned the following value: %p (ignored)", th_ret);
	}
	
	/* Clean the location */
	*th = (pthread_t)NULL;
	
	return ret;
}



