/*********************************************************************************************************
* Software License Agreement (BSD License)                                                               *
* Author: Sebastien Decugis <sdecugis@nict.go.jp>							 *
*													 *
* Copyright (c) 2008, WIDE Project and NICT								 *
* All rights reserved.											 *
* 													 *
* Redistribution and use of this software in source and binary forms, with or without modification, are  *
* permitted provided that the following conditions are met:						 *
* 													 *
* * Redistributions of source code must retain the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer.										 *
*    													 *
* * Redistributions in binary form must reproduce the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer in the documentation and/or other						 *
*   materials provided with the distribution.								 *
* 													 *
* * Neither the name of the WIDE Project or NICT nor the 						 *
*   names of its contributors may be used to endorse or 						 *
*   promote products derived from this software without 						 *
*   specific prior written permission of WIDE Project and 						 *
*   NICT.												 *
* 													 *
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED *
* WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A *
* PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR *
* ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 	 *
* LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 	 *
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR *
* TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF   *
* ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.								 *
*********************************************************************************************************/

/* Manage the list of RADIUS clients, along with their shared secrets. */

#include "radius_gw_internal.h"

/* Ordered lists of clients. The order relationship is a memcmp on the address zone. 
   For same addresses, the port is compared.
   A given address cannot be added with a 0-port and another port value.
 */
static struct rg_list cli_ip, cli_ip6;

/* Mutex to protect the previous lists */
static pthread_mutex_t cli_mtx = PTHREAD_MUTEX_INITIALIZER;

/* Structure describing one client */
struct cli_info {
	struct rg_list		chain;
	union {
		struct sockaddr_in	*sin;
		struct sockaddr_in6	*sin6;
	};
	struct {
		unsigned char * data;
		size_t		len;
	} 			key;
	int			refcount;
	
	/* Previous msg received, for duplicate checks. [0] for auth, [1] for acct. */
	struct {
		uint16_t	port;
		uint8_t		id;
	} last[2];
};

/* Function to alloc a new cli_info */
static int cli_info_create(struct cli_info ** res, struct sockaddr ** ip_port, unsigned char ** key, size_t keylen )
{
	struct cli_info *tmp = NULL;
	
	/* Create the new object */
	CHECK_MALLOC( tmp = malloc(sizeof (struct cli_info)) );
	memset(tmp, 0, sizeof(struct cli_info));
	
	/* Initialize the chain */
	rg_list_init(&tmp->chain);
	
	/* move the sa info reference */
	if ((*ip_port)->sa_family == AF_INET)
		tmp->sin = (struct sockaddr_in *) (*ip_port);
	else
		tmp->sin6 = (struct sockaddr_in6 *) (*ip_port);
	*ip_port = NULL;
	
	/* move the key material */
	tmp->key.data = *key;
	tmp->key.len = keylen;
	*key = NULL;
	
	/* Done! */
	*res = tmp;
	return 0;
}


/* Decrease refcount on a client; the lock must be held when this function is called. */
static void cli_info_unlink(struct cli_info * client)
{
	client->refcount -= 1;
	
	if (client->refcount <= 0) {
		/* to be sure */
		ASSERT( rg_list_is_empty(&client->chain) );
		
		/* Free the data */
		free(client->sin);
		free(client->key.data);
		free(client);
	}
}


/* Function to look for an existing cli_info, or the previous element. 
   The cli_mtx must be held when calling this func. 
   Returns ENOENT if the matching client does not exist, and res points to the previous element in the list. 
   Returns EEXIST if the matching client is found, and res points to this element. 
   Returns other error code on other error. */
static int cli_info_search(struct cli_info ** res, struct sockaddr * ip_port )
{
	int ret = 0;
	int cmp;
	struct rg_list *ref = NULL;
	
	CHECK_PARAMS(res && ip_port);
	
	switch (ip_port->sa_family) {
		case AF_INET: {
			struct sockaddr_in * sin = (struct sockaddr_in *)ip_port;
			
			for (ref = cli_ip.next; ref != &cli_ip; ref = ref->next) {
				cmp = memcmp(&sin->sin_addr, &((struct cli_info *)ref)->sin->sin_addr, sizeof(struct in_addr));
				
				if (cmp > 0 )
					continue; /* we must search further in the list */
				
				if (cmp < 0 ) /* This IP is not in the list, we return the previous element */
					break;
				
				/* Now we found an entry with the same IP; let's check the port number if both are present */
				
				/* If the ip_port we are searching does not contain a port, just return the first match result */
				/* If the entry in the list does not contain a port, return it as a match */
				/* If both ports are equal, it is a match */
				if ((sin->sin_port == 0) 
						|| (((struct cli_info *)ref)->sin->sin_port == 0) 
						|| (sin->sin_port == ((struct cli_info *)ref)->sin->sin_port)) {
					*res = (struct cli_info *)ref;
					return EEXIST;
				}
				
				/* now we order on the port value (we don't care aboit the network or host byte order there */
				if (sin->sin_port > ((struct cli_info *)ref)->sin->sin_port)
					continue;
				else
					break;
			}
			
			*res = (struct cli_info *)(ref->prev);
			return ENOENT;
		}
		break;
		
		case AF_INET6: {
			struct sockaddr_in6 * sin6 = (struct sockaddr_in6 *)ip_port;
			
			/* Exact same logic as for IP */
			for (ref = cli_ip6.next; ref != &cli_ip6; ref = ref->next) {
				cmp = memcmp(&sin6->sin6_addr, &((struct cli_info *)ref)->sin6->sin6_addr, sizeof(struct in6_addr));
				if (cmp > 0 )
					continue;
				if (cmp < 0 )
					break;
				if ((sin6->sin6_port == 0) 
						|| (((struct cli_info *)ref)->sin6->sin6_port == 0) 
						|| (sin6->sin6_port == ((struct cli_info *)ref)->sin6->sin6_port)) {
					*res = (struct cli_info *)ref;
					return EEXIST;
				}
				if (sin6->sin6_port > ((struct cli_info *)ref)->sin6->sin6_port)
					continue;
				else
					break;
			}
			*res = (struct cli_info *)(ref->prev);
			return ENOENT;
		}
	}
	
	/* We're never supposed to reach this point */
	ASSERT(0);
	return EINVAL;
}

static void cli_key_dump(char * keydump /* 60 bytes */, char * key, size_t keylen)
{
	int i, j, idx;
	char * ptr;

	memset(keydump, 0, 60);
	idx = 0;

	for (i = 0; i < 2; i++) {
		for (j = 0; j < 8; j++) {
			if (idx >= keylen)
				break;

			if (j == 4) {
				*keydump = ' '; keydump++;
			}

			sprintf(keydump, "%02hhX ", key[idx]);
			keydump += 3;
			idx ++;
		}
		
		*keydump = '\t'; keydump++;
	}
	if (keylen > idx) {
		sprintf(keydump, "...");
	}
}

int rgw_clients_getkey(void * cli, unsigned char **key, size_t *key_len)
{
	struct cli_info * cliinfo = (struct cli_info *) cli;
	
	CHECK_PARAMS( cli && key && key_len );
	*key = cliinfo->key.data;
	*key_len = cliinfo->key.len;
	return 0;
}

int rgw_clients_search(struct sockaddr * ip_port, void ** ref)
{
	int ret = 0;
	struct cli_info * cli;
	
	TRACE_ENTRY("%p %p", ip_port, ref);
	
	CHECK_PARAMS(ip_port && ref);
	*ref = NULL;
	
	CHECK_POSIX( pthread_mutex_lock(&cli_mtx) );
	ret = cli_info_search(&cli, ip_port);
	
	if (ret == EEXIST) {
		cli->refcount ++;
		ret = 0;
		*ref = cli;
	}
	
	CHECK_POSIX( pthread_mutex_unlock(&cli_mtx) );
	
	return ret;
}

int rgw_clients_checkdup(rad_t **msg, void *cli)
{
	struct cli_info * cli_info = (struct cli_info *) cli;
	int idx;
	
	TRACE_ENTRY("%p %p", msg, cli);
	
	CHECK_PARAMS( msg && cli );
	
	if ((*msg)->flags.serv_type == RGW_EXT_PORT_AUTH)
		idx = 0;
	else
		idx = 1;
	
	if ((cli_info->last[idx].id == (*msg)->identifier) && (cli_info->last[idx].port == (*msg)->port)) {
		/* Duplicate! */
		TRACE_DEBUG(INFO, "Received duplicated RADIUS message (id: %02hhx, port: %hu), discarding.", (*msg)->identifier, (*msg)->port);
		rg_msg_free(*msg);
		*msg = NULL;
	} else {
		/* Save new information */
		cli_info->last[idx].id = (*msg)->identifier;
		cli_info->last[idx].port = (*msg)->port;
	}
	
	return 0;
}

void rgw_clients_dispose(void ** ref)
{
	TRACE_ENTRY("%p", ref);
	CHECK_PARAMS_DO(ref, return);
	
	CHECK_POSIX_DO( pthread_mutex_lock(&cli_mtx),  );
	cli_info_unlink(*(struct cli_info **)ref);
	*ref = NULL;
	CHECK_POSIX_DO( pthread_mutex_unlock(&cli_mtx), );
}

int rgw_clients_init(void)
{
	rg_list_init(&cli_ip);
	rg_list_init(&cli_ip6);
	return 0;
}

int rgw_clients_add( struct sockaddr * ip_port, unsigned char ** key, size_t keylen )
{
	struct cli_info * prev = NULL, *new = NULL;
	int ret;
	
	TRACE_ENTRY("%p %p %lu", ip_port, key, keylen);
	
	CHECK_PARAMS( ip_port && key && *key && keylen );
	CHECK_PARAMS( (ip_port->sa_family == AF_INET) || (ip_port->sa_family == AF_INET6) );
	
	/* Dump the entry in debug mode */
	if (TRACE_BOOL(FULL + 1 )) {
		char ipstr[INET6_ADDRSTRLEN];
		char keydump[60];
		uint16_t port;
		
		if (ip_port->sa_family == AF_INET) {
			inet_ntop(AF_INET, &((struct sockaddr_in *)ip_port)->sin_addr,ipstr,sizeof(ipstr));
			port = ntohs(((struct sockaddr_in *)ip_port)->sin_port);
		} else {
			inet_ntop(AF_INET6, &((struct sockaddr_in6 *)ip_port)->sin6_addr,ipstr,sizeof(ipstr));
			port = ntohs(((struct sockaddr_in6 *)ip_port)->sin6_port);
		}
		
		cli_key_dump(&keydump[0], *key, keylen);
		
		TRACE_DEBUG(FULL, "Adding client [%s]:%hu with %d bytes key: %s", ipstr, port, keylen, keydump);
	}
	
	/* Lock the lists */
	CHECK_POSIX( pthread_mutex_lock(&cli_mtx) );
	
	/* Check if the same entry does not already exist */
	ret = cli_info_search(&prev, ip_port );
	if (ret == ENOENT) {
		CHECK_FCT_DO( ret = cli_info_create( &new, &ip_port, key, keylen ), goto end );
		
		rg_list_insert_after(&prev->chain, &new->chain);
		
		new->refcount++;
		
		ret = 0;
		goto end;
	}
	
	if (ret == EEXIST) {
		/* Check if the key is the same, then skip or return an error */
		if ((keylen == prev->key.len ) && ( ! memcmp(*key, prev->key.data, keylen) )) {
			TRACE_DEBUG(INFO, "Skipping duplicate client description");
			goto end;
		}
		
		TRACE_DEBUG(INFO, "Error: conflicting entries");	
		log_error("Error adding a RADIUS client: conflict with a previous entry.\n");
		{
			char ipstr[INET6_ADDRSTRLEN];
			char keydump[60];
			uint16_t port;

			if (prev->sin->sin_family == AF_INET) {
				inet_ntop(AF_INET, &prev->sin->sin_addr,ipstr,sizeof(ipstr));
				port = ntohs(prev->sin->sin_port);
			} else {
				inet_ntop(AF_INET6, &prev->sin6->sin6_addr,ipstr,sizeof(ipstr));
				port = ntohs(prev->sin6->sin6_port);
			}

			cli_key_dump(&keydump[0], prev->key.data, prev->key.len);

			log_error( "Previous entry: [%s]:%hu, key (%db): %s\n", ipstr, port, prev->key.len, keydump);
			
			if (ip_port->sa_family == AF_INET) {
				inet_ntop(AF_INET, &((struct sockaddr_in *)ip_port)->sin_addr,ipstr,sizeof(ipstr));
				port = ntohs(((struct sockaddr_in *)ip_port)->sin_port);
			} else {
				inet_ntop(AF_INET6, &((struct sockaddr_in6 *)ip_port)->sin6_addr,ipstr,sizeof(ipstr));
				port = ntohs(((struct sockaddr_in6 *)ip_port)->sin6_port);
			}

			cli_key_dump(&keydump[0], *key, keylen);
			log_error( "New entry:      [%s]:%hu, key (%db): %s\n", ipstr, port, keylen, keydump);			
		}
	}
end:
	/* Lock the lists */
	CHECK_POSIX( pthread_mutex_unlock(&cli_mtx) );
	
	return ret;
}

void rgw_clients_dump(void)
{
	struct cli_info * client = NULL;
	struct rg_list *ref = NULL;
	char ipstr[INET6_ADDRSTRLEN];
	char keydump[60];
	uint16_t port;
	
	if ( ! TRACE_BOOL(FULL) )
		return;
	
	CHECK_POSIX_DO( pthread_mutex_lock(&cli_mtx), /* ignore error */ );
	
	log_debug(" RADIUS IP clients list:\n");
	for (ref = cli_ip.next; ref != &cli_ip; ref = ref->next) {
		
		client = (struct cli_info *)ref;
		
		inet_ntop(AF_INET, &client->sin->sin_addr,ipstr,sizeof(ipstr));
		port = ntohs(client->sin->sin_port);
		
		cli_key_dump(&keydump[0], client->key.data, client->key.len);
		
		log_debug("   [%s]:%hu, %d bytes: %s\n", ipstr, port, client->key.len, keydump);
	}
		
	log_debug(" RADIUS IPv6 clients list:\n");
	for (ref = cli_ip6.next; ref != &cli_ip6; ref = ref->next) {
		
		client = (struct cli_info *)ref;
		
		inet_ntop(AF_INET6, &client->sin6->sin6_addr,ipstr,sizeof(ipstr));
		port = ntohs(client->sin6->sin6_port);
		
		cli_key_dump(&keydump[0], client->key.data, client->key.len);
		
		log_debug("   [%s]:%hu, %d bytes: %s\n", ipstr, port, client->key.len, keydump);
	}
		
	CHECK_POSIX_DO( pthread_mutex_unlock(&cli_mtx), /* ignore error */ );
}

void rgw_clients_fini(void)
{
	struct rg_list * client;
	
	TRACE_ENTRY();
	
	CHECK_POSIX_DO( pthread_mutex_lock(&cli_mtx), /* ignore error */ );
	
	/* empty the lists */
	while ( ! rg_list_is_empty(&cli_ip) ) {
		client = cli_ip.next;
		rg_list_unlink(client);
		cli_info_unlink((struct cli_info *)client);
	}
	while (! rg_list_is_empty(&cli_ip6)) {
		client = cli_ip6.next;
		rg_list_unlink(client);
		cli_info_unlink((struct cli_info *)client);
	}
	
	CHECK_POSIX_DO( pthread_mutex_unlock(&cli_mtx), /* ignore error */ );
	
}

