/*********************************************************************************************************
* Software License Agreement (BSD License)                                                               *
* Author: Sebastien Decugis <sdecugis@nict.go.jp>							 *
*													 *
* Copyright (c) 2008, WIDE Project and NICT								 *
* All rights reserved.											 *
* 													 *
* Redistribution and use of this software in source and binary forms, with or without modification, are  *
* permitted provided that the following conditions are met:						 *
* 													 *
* * Redistributions of source code must retain the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer.										 *
*    													 *
* * Redistributions in binary form must reproduce the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer in the documentation and/or other						 *
*   materials provided with the distribution.								 *
* 													 *
* * Neither the name of the WIDE Project or NICT nor the 						 *
*   names of its contributors may be used to endorse or 						 *
*   promote products derived from this software without 						 *
*   specific prior written permission of WIDE Project and 						 *
*   NICT.												 *
* 													 *
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED *
* WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A *
* PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR *
* ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 	 *
* LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 	 *
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR *
* TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF   *
* ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.								 *
*********************************************************************************************************/

/* Manage incoming RADIUS message. */

#include "radius_gw_internal.h"

/* How many threads to work on messages ? */
#define NB_WORKERS	2


static pthread_mutex_t work_mtx = PTHREAD_MUTEX_INITIALIZER;
static pthread_cond_t work_cond = PTHREAD_COND_INITIALIZER;

static struct rg_list work_data;
struct work_item {
	struct rg_list chain;
	rad_t * msg;
	void * cli;
};

struct pending_answer {
	rad_t     * rad;
	void      * cli;
	sess_id_t * sess;
};

static sess_reg_t * sess_hdl = NULL;

static pthread_t workers[NB_WORKERS];

static void cleanup_release_mtx(void * mtx)
{
	CHECK_POSIX_DO( pthread_mutex_unlock((pthread_mutex_t *)mtx), );
}

/* This function is strongly inspired from radius_msg_verify_msg_auth in hostap package (BSD licensed) */
static int work_auth_check(rad_t * rad, void * cli, unsigned char * req_auth)
{
	unsigned char buf[RADIUS_MAX_MSG_LEN];
	unsigned char * key;
	size_t key_len;
	struct rg_list * attr_li;
	struct rad_attr * msg_mac = NULL;
	unsigned char msg_mac_save[MD5_MAC_LEN];
	unsigned char msg_mac_calc[MD5_MAC_LEN];
	unsigned char msg_auth_orig[16];
	size_t buflen = RADIUS_MAX_MSG_LEN;
	
	TRACE_ENTRY("%p %p", rad, cli);
	
	/* Find the authenticator attribute */
	for (attr_li = rad->attributes.next; attr_li != &rad->attributes; attr_li = attr_li->next) {
		struct rad_attr * attr = (struct rad_attr *)attr_li;
		
		if (attr->type == RADIUS_ATTR_MESSAGE_AUTHENTICATOR) {
			if (msg_mac != NULL) {
				TRACE_DEBUG(INFO, "RADIUS message with several Message-Authenticator attributes, invalid.");
				return EINVAL;
			}
			msg_mac = attr;
		}
	}
	if (msg_mac == NULL) {
		TRACE_DEBUG(FULL, "RADIUS message without Message-Authenticator attribute.");
		return 0;
	}
	if (msg_mac->length != (2 + sizeof(msg_mac_save))) {
		TRACE_DEBUG(INFO, "Unexpected size of Message-Authenticator attribute, discarding.");
		return EINVAL;
	}
	
	/* Save and clear the authenticator value */
	memcpy( &msg_mac_save[0], &msg_mac->data.buf[0], sizeof(msg_mac_save) );
	memset( &msg_mac->data.buf[0], 0, sizeof(msg_mac_save) );
	if (req_auth) {
		/* compute the MAC with the authenticator from the request */
		memcpy( &msg_auth_orig[0], &rad->authenticator[0], 16 );
		memcpy( &rad->authenticator[0], req_auth, 16 );
	}
	
	/* Recreate a false message to compute the md5 sum. This is very bad design, optimization possible here... */
	memset(buf, 0, sizeof(buf));
	key = &buf[0]; /* use key here to avoid declarataion of another var */
	CHECK_FCT( rgw_msg_gen(rad, &key, &buflen) );
	
	/* Get the shared secret associated with this client */
	CHECK_FCT( rgw_clients_getkey(cli, &key, &key_len) );
	
	/* Now compute the MAC with our shared key */
	hmac_md5( key, key_len, &buf[0], buflen, &msg_mac_calc[0] );
	
	/* Restore the original msg */
	memcpy( &msg_mac->data.buf[0], &msg_mac_save[0], sizeof(msg_mac_save) );
	if (req_auth) {
		/* also restore the authenticator */
		memcpy( &rad->authenticator[0], &msg_auth_orig[0], 16 );
	}
	
	/* Finally, check if the MAC is correct (validates the common shared secret) */
	if (memcmp(&msg_mac_save[0], &msg_mac_calc[0], sizeof(msg_mac_save) )) {
		TRACE_DEBUG(INFO, "Invalid Message-Authenticator attribute (bad shared secret?).");
		return EINVAL;
	} else {
		rad->flags.auth_ok = 1;
		TRACE_DEBUG(FULL, "Valid Message-Authenticator attribute found.");
	}
	
	return 0;
}


static void work_receive_diam_answer(void * paback, msg_t **ans)
{
	struct pending_answer * pa = (struct pending_answer *)paback;
	
	TRACE_ENTRY("%p %p", pa, ans);
	
	TRACE_DEBUG(INFO, "Handling Diameter answer: Not implemented yet...");
	
}

static void * work_th(void * arg)
{
	char thname[10];
	
	TRACE_ENTRY("%p", arg);
	
	snprintf(thname, sizeof(thname), "worker #%d", (int)arg);
	THREAD_NAME(thname);
	
	while (1) { /* The thread will be cancelled */
		
		rad_t * msg;
		void * cli;
		struct work_item * wi;
		struct rg_list * li;
		sess_id_t * session;
		msg_t * diam_msg;
		int pb;
		struct pending_answer * pa;
	
		/* Pick the next message */
		CHECK_POSIX_DO( pthread_mutex_lock(&work_mtx), return NULL );
		
		pthread_cleanup_push(cleanup_release_mtx, &work_mtx);

		/* wait for new work data */
		while (rg_list_is_empty(&work_data)) {
			CHECK_POSIX_DO( pthread_cond_wait( &work_cond, &work_mtx ), 
				{
					pthread_mutex_unlock(&work_mtx);
					return NULL;
				}  );
		}
		
		/* Retrieve the next element */
		wi = (struct work_item *)(work_data.next);
		rg_list_unlink(&wi->chain);
		msg = wi->msg;
		cli = wi->cli;
		free(wi);
		
		/* Release the mutex */
		pthread_cleanup_pop(0);
		CHECK_POSIX_DO( pthread_mutex_unlock(&work_mtx), return NULL );
		
		TRACE_DEBUG(FULL, "Processing next message: %p received on client: %p", msg, cli);
	
		/* process the data */
		
		/* Check authenticator */
		CHECK_FCT_DO( work_auth_check(msg, cli, NULL),
			{
				/* An error occurred, discard message */
				rg_msg_free(msg);
				rgw_clients_dispose(&cli);
				continue;
			}  );
		
		/* Check duplicate */
		CHECK_FCT_DO( rgw_clients_checkdup(&msg, cli),
			{
				/* An error occurred, discard message */
				rg_msg_free(msg);
				rgw_clients_dispose(&cli);
				continue;
			}  );
		if (msg == NULL) {
			rgw_clients_dispose(&cli);
			continue; /* the message was a duplicate */
		}
		
		/* Pass the message to the list of registered extensions */
		session = NULL;
		diam_msg = NULL;
		CHECK_FCT_DO( rgw_extensions_loop_req(&msg, &session, &diam_msg, cli), 
			{
				/* An error occurred, discard message */
				rg_msg_free(msg);
				rgw_clients_dispose(&cli);
				continue;
			}  );
		if (msg == NULL) {
			rgw_clients_dispose(&cli);
			continue; /* the message was a duplicate */
		}
		
		/* Check if the full content of the RADIUS message was handled */
		pb = 0;
		for (li = msg->attributes.next; li != &msg->attributes; li = li->next) {
			struct rad_attr * attr = (struct rad_attr *)li;
			if (! attr->flags.handled) {
				pb++;
				log_error("No extension available to handle attribute %hhd (%s) in command %hhd (%s)!\n",
						attr->type, rg_msg_attrtype_str(attr->type),
						msg->code, rg_msg_code_str(msg->code));
			}
		}
		
		/* Check the created Diameter message */
		if ((diam_msg == NULL) || ( msg_parse_rules(diam_msg, NULL) ) ) {
			log_error("No or invalid Diameter generated from RADIUS command %hhd (%s).\n"
					" Turn on advanced log for detail.\n",
					msg->code, rg_msg_code_str(msg->code));
			/* We might also dump the conflicting rule here if useful */
			pb++;
		}
		
		/* Check the session is correct */
		if (session == NULL) {
			log_error("No session has been created to store RADIUS state (command %hhd (%s)).\n"
					" Please check your configuration and documentation.\n",
					msg->code, rg_msg_code_str(msg->code));
			pb++;
		}
		
		if (pb) {
			/* Something went wrong during the conversion */
			if (session) {
				CHECK_FCT_DO( sess_unlink(session), );
				session = NULL;
			}
			
			if (diam_msg) {
				CHECK_FCT_DO( msg_free(diam_msg, 1), );
				diam_msg = NULL;
			}
			
			rg_msg_free(msg);
			rgw_clients_dispose(&cli);
			
			TRACE_DEBUG(INFO, "A problem occurred while translating a RADIUS message, data discarded.\n");
			continue;
		}
				
		/* Send the radius message and register for answer */
		CHECK_MALLOC_DO( pa = malloc(sizeof(struct pending_answer)), break );
		memset(pa, 0, sizeof(*pa));
		pa->rad = msg;
		pa->cli = cli;
		pa->sess= session;
		
		CHECK_FCT_DO( msg_send( &diam_msg, work_receive_diam_answer, pa), 
			{
				/* If an error occurs, log and destroy the data */
				log_error("An error occurred while sending Diameter message, please turn Debug on for detail.\n");
				if (session) {
					CHECK_FCT_DO( sess_unlink(session), );
					session = NULL;
				}

				if (diam_msg) {
					CHECK_FCT_DO( msg_free(diam_msg, 1), );
					diam_msg = NULL;
				}

				rg_msg_free(msg);
				rgw_clients_dispose(&cli);
				
				free(pa);
				
				continue;
			} );
		
		/* Done! */
	}
	
	TRACE_DEBUG(INFO, "Error: thread terminated!");
	return NULL;
}


int rgw_work_start(void)
{
	int i;
	TRACE_ENTRY();
	
	CHECK_FCT( sess_regext(&sess_hdl) );
	
	memset(workers, 0, sizeof(workers));
	rg_list_init(&work_data);
	
	/* Create the worker thread(s) */
	for (i = 0; i < NB_WORKERS; i++) {
		CHECK_POSIX( pthread_create(&workers[i], NULL, work_th, (void *)i) );
	}
	
	return 0;
}

int rgw_work_add(rad_t * msg, void * client)
{
	struct work_item * new;
	
	CHECK_MALLOC( new = malloc(sizeof(struct work_item)) );
	memset(new, 0, sizeof(struct work_item));
	rg_list_init(&new->chain);
	
	new->msg = msg;
	new->cli = client;
	
	CHECK_POSIX( pthread_mutex_lock(&work_mtx) );
	rg_list_insert_before(&work_data, &new->chain);
	CHECK_POSIX( pthread_cond_signal(&work_cond) );
	CHECK_POSIX( pthread_mutex_unlock(&work_mtx) );
	
	return 0;
}

void rgw_work_fini(void)
{
	int i;
	TRACE_ENTRY();
	
	for (i = 0; i < NB_WORKERS; i++) {
		rg_thread_term(&workers[i]);
	}
	
	CHECK_FCT_DO( sess_deregext(sess_hdl), );
		
	return;
}
