/*********************************************************************************************************
* Software License Agreement (BSD License)                                                               *
* Author: Sebastien Decugis <sdecugis@nict.go.jp>							 *
*													 *
* Copyright (c) 2008, WIDE Project and NICT								 *
* All rights reserved.											 *
* 													 *
* Redistribution and use of this software in source and binary forms, with or without modification, are  *
* permitted provided that the following conditions are met:						 *
* 													 *
* * Redistributions of source code must retain the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer.										 *
*    													 *
* * Redistributions in binary form must reproduce the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer in the documentation and/or other						 *
*   materials provided with the distribution.								 *
* 													 *
* * Neither the name of the WIDE Project or NICT nor the 						 *
*   names of its contributors may be used to endorse or 						 *
*   promote products derived from this software without 						 *
*   specific prior written permission of WIDE Project and 						 *
*   NICT.												 *
* 													 *
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED *
* WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A *
* PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR *
* ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 	 *
* LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 	 *
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR *
* TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF   *
* ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.								 *
*********************************************************************************************************/

/* This file contains the definitions of types and functions related to the peers facility
 and that can be called by extensions. */
 
 
#ifndef _PEER_API_H
#define _PEER_API_H

#include <stddef.h>
#include <time.h>

#include <waaad/dictionary-api.h>

/*
 ***************************************************************************
 *
 * Peer API functions 
 *
 * These functions are used to manage the collection of peers to which
 * the local host is directly connected.
 *
 ***************************************************************************
 */
 
/* This following opaque type is used to represent a peer object. */
typedef void peer_t;

/* This type defines an applications supported by a peer (used as array). The elements are always ordered by their 'a' value. Last element is 0. */
typedef enum { APP_TYPE_AUTH, APP_TYPE_ACCT } peer_app_type_t;
typedef struct {
	application_id_t a;
	vendor_id_t	 v; /* 0 for non-Vendor-Specific applications */
	peer_app_type_t	 t;
} peer_appl_t;	

/* This enum represents the fields that can be required with peer_get */
typedef enum {
	/* The Diameter-Id of this peer */
	PEER_DIAMETER_ID = 1,	/* *result is a (char *) pointing to the peer Diameter Id */
	
	/* If you are lazy to compute it from the diameter-id */
	PEER_REALM,		/* *result is a (char *) pointing to the peer realm */
			
	/* The state of this peer */
	PEER_ONLINE,		/* *result is 1 if the peer is in "open" state, 0 otherwise */
			
	/* List of applications supported by this peer */
	PEER_APPLICATIONS	/* *result is a pointer to a peer_appl_t array. */

} peer_info_t;

/* The following type is used to create a new peer */
typedef struct _pa_ {
	
	/* Flags: */
#define PA_NOSEC	(1<<0)	/* disable completly the security ? -- SHOULD BE DEPRECATED!!! */
#define PA_EXPIRY	(1<<1)	/* for peers dynamically discovered, set this flag and the lifetime in pa_ts (DNS TTL, ...). if pa_ts is 0, the peer expires at transport disconnection */
#define PA_CB		(1<<2)	/* the pa_cb callback must be called with the pa_cbparm as parameter when peer expires (see pa_cb definition) */
#define PA_FREE		(1<<3)	/* when peer is destroyed, free() must be called on the pa_cbparm */
	uint32_t	pa_flags;	/* see previous flags definitions */
	
	/* Expiration data */
	struct timespec	pa_ts;		/* the lifetime of the peer (absolute timespec of expiry). */
	int 		(*pa_cb)(void *, struct _pa_ *);
					/* the pa_cb callback is called when peer expires and PA_CB flag is set.
					  It receives the pointer to this structure as parameter.
					  It should update the pa_ts information.
					  It may also update the other fields, for example pa_ss.
					  It must return:
					     0 if the connection is still valid and timeout as been updated.
					    >0 if timeout and connection info were updated, and the connection must be reset.
					    <0 if the peer must be disconnected and deleted.
					  The next field pa_cbparm is not used by waaad core and can be used to pass parameters to pa_cb,
					  such as a SRV name to resolv for example.
					 */
	void		*pa_cbparm;   

	/* Connection information */
	int		pa_proto;	/* IPPROTO_TCP or IPPROTO_SCTP */
	sSS		pa_ss;		/* the primary information (IP address and port) to connect to the peer */
	uint16_t	pa_streams;	/* number of streams for SCTP, ignored for TCP */
	
} peer_add_t;

#ifndef IN_EXTENSION


/*
 * FUNCTION:	peer_add
 *
 * PARAMETERS:
 *  diamid	: the diameter id of the peer (will be converted to lower-case and copied).
 *  len		: the length of the diameter id string (which do not need to be \0-terminated)
 *  data 	: Information to create the peer.
 *
 * DESCRIPTION: 
 *  Add a peer to the list of peers to which the daemon must maintain a connexion.
 *  The values are copied inside the module and can be disposed after the call.
 *
 * RETURN VALUE:
 *  0      	: The peer is added.
 *  EINVAL 	: A parameter is invalid.
 *  EEXIST 	: A peer with the same Diameter-Id is already in the list.
 *  (other standard errors may be returned, too, with their standard meaning. Example:
 *    ENOMEM 	: Memory allocation for the new object element failed.)
 */
int peer_add ( char * diamid, size_t len, peer_add_t * data );

/*
 * FUNCTION:	peer_get
 *
 * PARAMETERS:
 *  peer 	: pointer to a valid peer object.
 *  what	: What information is required?
 *  result	: upon success, a pointer to the requested data is stored here.
 *		 the pointed data must not be modified.
 *		 See peer_info_t description for more information on the returned pointer.
 *
 * DESCRIPTION: 
 *  Get information about a peer. Note that the result pointer points to internal data that must
 * not be modified by calling app.
 *
 * RETURN VALUE:
 *  0      	: The peer is added.
 *  EINVAL 	: A parameter is invalid.
 *  EEXIST 	: A peer with the same Diameter-Id is already in the list.
 *  (other standard errors may be returned, too, with their standard meaning. Example:
 *    ENOMEM 	: Memory allocation for the new object element failed.)
 */
int peer_get ( peer_t * peer, peer_info_t what, void **result );

#endif /* ! IN_EXTENSION */

/* The version of this API, to check binary compatibility -- increment each time a change is made in api_peer_t */
#define WAAAD_API_PEER_VER	1

/* Now define the type of the structure that contains the callback to pass to extensions */
typedef struct {
	/* The header is common to all sub-API pieces */
	size_t	length;		/* The size of this structure, may be useful for extensions not using the facility */
	int	version;	/* The version of this API/ABI, must be WAAAD_API_PEER_VER */
	
	/* the remaining is peer-specific */
	int  (*peer_add)  ( char * diamid, size_t len, peer_add_t * data );
	int  (*peer_get)  ( peer_t * peer, peer_info_t what, void **result );
} api_peer_t;

#ifdef IN_EXTENSION

/* From within the extensions, we register callbacks in the following global structure */
#ifdef DECLARE_API_POINTERS
api_peer_t * g_api_peer=NULL;
#else /* DECLARE_API_POINTERS */
extern api_peer_t * g_api_peer;
#endif /* DECLARE_API_POINTERS */

/* These defines allow to call functions from extension in the same way as in waaad */
#define peer_add	g_api_peer->peer_add
#define peer_get	g_api_peer->peer_get


#else /* IN_EXTENSION */

/* From the daemon, we must initialize the API object, in extension.c */
# define MY_WAAAD_API_PEER_VER 1
# if MY_WAAAD_API_PEER_VER != WAAAD_API_PEER_VER
#  error "You must update API_INIT_PEER also"
# endif

#define API_INIT_PEER( api_peer ) 			\
{							\
	(api_peer).length     = sizeof(api_peer_t);	\
	(api_peer).version    = WAAAD_API_PEER_VER;	\
	(api_peer).peer_add   = peer_add;		\
	(api_peer).peer_get   = peer_get;		\
}							

#endif /* IN_EXTENSION */

#endif /* ! _PEER_API_H */
